from datetime import datetime, date, time
import math
from ...models.hourselection.hourselection_options import HourSelectionOptions
from .schedule_session import ScheduleSession


class Scheduler:
    """This class obj is what constitutes a running scheduler."""
    def __init__(self, options:HourSelectionOptions = None):
        self.model = ScheduleSession(hourselection_options=options)
        self.active = False

    @property
    def scheduler_active(self) -> bool:
        if self.active is False:
            return False
        return self.model.departuretime > datetime.now() or self.model.remaining_charge > 0

    # def create(
    #     self, 
    #     desired_charge:float, 
    #     departuretime:datetime, 
    #     starttime:datetime=datetime.now(),
    #     override_settings = False
    #     ):
    #     if self.scheduler_active:
    #         self._cancel()
    #     self.model.departuretime = departuretime
    #     self.model.starttime = starttime
    #     self.model.remaining_charge = desired_charge
    #     self.model._override_settings = override_settings
    
    # def _update(
    #     self,
    #     avg24:float,
    #     peak:float,
    #     charged_amount:float = None,
    #     prices:list = None,
    #     prices_tomorrow:list = None,
    #     mockdt:datetime = datetime.now()
    #     ):
    #     """calculate based on the pricing of hours, current peak and the avg24hr energy consumption"""
    #     self.model.MOCKDT = mockdt
    #     self.active = True
    #     if self.model.remaining_charge <= 0 or self.model.departuretime <= mockdt:
    #         return self._cancel()
    #     charge_per_hour = peak - (avg24/1000)
    #     if charge_per_hour <= 0:
    #         raise Exception
        
    #     self.model.remaining_charge -= charged_amount if charged_amount is not None else 0
    #     self.model.hours_price = [prices, prices_tomorrow]
    #     cheapest = self._sort_pricelist()
    #     self.model.hours_charge = self._get_charge_hours(
    #         cheapest_hours=cheapest, 
    #         charge_per_hour=charge_per_hour,
    #         peak=peak
    #         )
    
    # def _cancel(self):
    #     self.active = False
    #     self.model.departuretime = datetime.min
    #     self.model.starttime = datetime.min
    #     self.model.remaining_charge = 0

    # def _sort_pricelist(self) -> dict:
    #     if self.model._override_settings is False and self.model.hourselection_options is not None:
    #         return self._filter_pricelist()    
    #     return dict(sorted(self.model.hours_price.items(), key=lambda item: item[1]))
    
    # def _filter_pricelist(self) -> dict:
    #     filtered = {key:value for (key,value) in self.model.hours_price.items() if value <= self.model.hourselection_options.absolute_top_price}
    #     ret= dict(sorted(filtered.items(), key=lambda item: item[1]))
    #     return ret

    # def _get_charge_hours(
    #     self, 
    #     cheapest_hours:dict, 
    #     charge_per_hour:float,
    #     peak:float
    #     ) -> dict:
    #     remainder = self.model.remaining_charge
    #     chargehours = dict()
    #     for c in cheapest_hours.keys():
    #         if remainder <= 0:
    #             break
    #         if remainder > charge_per_hour:
    #             chargehours[c] = 1
    #         elif 0 < remainder < charge_per_hour:
    #             chargehours[c] = math.ceil((remainder/peak)*10)/10
    #         remainder -= charge_per_hour
    #     return chargehours

    async def async_create(self, desired_charge:float, departuretime:datetime, starttime:datetime=datetime.now(), override_settings = False):
        if self.scheduler_active:
            await self.async_cancel()
        self.model.departuretime = departuretime
        self.model.starttime = starttime
        self.model.remaining_charge = desired_charge
        self.model._override_settings = override_settings

    async def async_update(
        self,
        avg24:float,
        peak:float,
        charged_amount:float = None,
        prices:list = None,
        prices_tomorrow:list = None,
        mockdt:datetime = datetime.now()
        ):
        """calculate based on the pricing of hours, current peak and the avg24hr energy consumption"""
        self.model.MOCKDT = mockdt
        self.active = True
        if self.model.remaining_charge <= 0 or self.model.departuretime <= mockdt:
            return await self.async_cancel()
        charge_per_hour = peak - (avg24/1000)
        if charge_per_hour <= 0:
            raise Exception
        
        self.model.remaining_charge -= charged_amount if charged_amount is not None else 0
        self.model.hours_price = [prices, prices_tomorrow]
        cheapest = await self.async_sort_pricelist()
        self.model.hours_charge = await self.async_get_charge_hours(
            cheapest_hours=cheapest, 
            charge_per_hour=charge_per_hour,
            peak=peak
            )

    async def async_cancel(self):
        self.active = False
        self.model.departuretime = datetime.min
        self.model.starttime = datetime.min
        self.model.remaining_charge = 0
    
    async def async_sort_pricelist(self) -> dict:
        if self.model._override_settings is False and self.model.hourselection_options is not None:
            return await self.async_filter_pricelist()    
        return dict(sorted(self.model.hours_price.items(), key=lambda item: item[1]))
    
    async def async_filter_pricelist(self) -> dict:
        filtered = {key:value for (key,value) in self.model.hours_price.items() if value <= self.model.hourselection_options.absolute_top_price}
        ret= dict(sorted(filtered.items(), key=lambda item: item[1]))
        return ret

    async def async_get_charge_hours(
        self, 
        cheapest_hours:dict, 
        charge_per_hour:float,
        peak:float
        ) -> dict:
        remainder = self.model.remaining_charge
        chargehours = dict()
        for c in cheapest_hours.keys():
            if remainder <= 0:
                break
            if remainder > charge_per_hour:
                chargehours[c] = 1
            elif 0 < remainder < charge_per_hour:
                chargehours[c] = math.ceil((remainder/peak)*10)/10
            remainder -= charge_per_hour
        return chargehours


class Schedule:
    start_date: date
    end_date: date
    start_time: time
    end_time: time
    charge_amount: float
    weekdays: int


class Schedules:
    #list of schedule with id
    pass

class SchedulerCommands:
    def __init__(self) -> None:
        pass

    def create(self) -> int:
        pass

    def edit(self, scheduler_id: int) -> bool:
        pass

    def delete(self, scheduler_id: int) -> bool:
        pass

    def pause(self, scheduler_id: int) -> bool:
        pass

    def resume(self, scheduler_id: int) -> bool:
        pass


from abc import abstractmethod
from ...util import _convert_quarterly_minutes
from datetime import datetime
import logging
from ...models.phases import Phases
from ...models.const import (
    CURRENTS_ONEPHASE_1_16, CURRENTS_THREEPHASE_1_16, CURRENTS_ONEPHASE_1_32, CURRENTS_THREEPHASE_1_32
)

_LOGGER = logging.getLogger(__name__)


ALGORITHM_INPUTS = {
    "stop_caution": [1.075, 0.0032, 0.7],
    "stop": [1.071, 0.00165, 0.8],
    "start_caution": [1.081, 0.0049, 0.4],
    "start": [1.066, 0.0045, 0.5]
}

def _calculate_algorithm_inputs(type:str, minute:int) -> float:
    inputs = ALGORITHM_INPUTS[type]
    return round((((minute+pow(inputs[0], minute)) * inputs[1]) + inputs[2]) * 100, 2)


class ThresholdBase:
    BASECURRENT = 6
    def __init__(self, hub):
        self._hub = hub
        self._phases = Phases.Unknown
        self._currents = {}

    @property
    def phases(self) -> str:
        return self._phases.name

    @property
    def currents(self) -> dict:
        return self._currents

    @property
    def stop(self) -> float:
        is_caution = str(datetime.now().hour) in self._hub.hours.caution_hours if self._hub.options.price.price_aware is False else False

        return ThresholdBase._stop(
            datetime.now().minute,
            is_caution,
            self._hub.sensors.locale.data.is_quarterly(self._hub.sensors.locale.data)
        )

    @property
    def start(self) -> float:
        is_caution = str(datetime.now().hour) in self._hub.hours.caution_hours if self._hub.options.price.price_aware is False else False

        return ThresholdBase._start(
            datetime.now().minute,
            is_caution,
            self._hub.sensors.locale.data.is_quarterly(self._hub.sensors.locale.data)
        )

    @property
    @abstractmethod
    def allowedcurrent(self) -> int:
        pass

    def _setcurrentdict(self) -> dict:
        """only allow amps if user has set this value high enough"""
        if self._hub.chargertype.max_amps != 16:
            _threephase = {k: v for (k, v) in CURRENTS_THREEPHASE_1_32.items() if v <= self._hub.chargertype.max_amps}
            _onephase = {k: v for (k, v) in CURRENTS_ONEPHASE_1_32.items() if v <= self._hub.chargertype.max_amps}
        else:
            _threephase = CURRENTS_THREEPHASE_1_16
            _onephase = CURRENTS_ONEPHASE_1_16
        if hasattr(self._hub.sensors, "carpowersensor"):
            try:
                divid = int(self._hub.sensors.carpowersensor.value)/int(self._hub.sensors.chargerobject_switch.current)
                if int(self._hub.sensors.carpowersensor.value) == 0:
                    self._phases = Phases.Unknown
                    ret = _threephase
                elif divid < 300:
                    self._phases = Phases.OnePhase
                    ret = _onephase
                else:
                    self._phases = Phases.ThreePhase
                    ret = _threephase
            except:
                _LOGGER.debug("Currents-dictionary: could not divide charger amps with charger power. Falling back to legacy-method.")
                if 0 < int(self._hub.sensors.carpowersensor.value) < 4000:
                    self._phases = Phases.OnePhase
                    ret = _onephase
                else:
                    self._phases = Phases.ThreePhase
                    ret = _threephase
        else:
            self._phases = Phases.ThreePhase
            ret = _threephase
        self._currents = ret
        return ret

    @staticmethod
    def _stop(
              now_min: int,
              is_caution_hour: bool,
              is_quarterly: bool=False
              ) -> float:
        minute = _convert_quarterly_minutes(now_min, is_quarterly) 
        if is_caution_hour and minute < 45:
            return _calculate_algorithm_inputs("stop_caution", minute)
        return _calculate_algorithm_inputs("stop", minute)

    @staticmethod
    def _start(
               now_min: int,
               is_caution_hour: bool,
               is_quarterly:bool=False
               ) -> float:
        minute = _convert_quarterly_minutes(now_min, is_quarterly)
        if is_caution_hour and minute < 45:
            return _calculate_algorithm_inputs("start_caution", minute)
        return _calculate_algorithm_inputs("start", minute)
    
    @staticmethod
    def allowed_current(
            now_min: int,
            moving_avg: float,
            charger_enabled: bool,
            charger_done: bool,
            currents_dict: dict,
            total_energy: float,
            peak: float,
            is_quarterly:bool=False,
            power_canary_amp: int = -1
            ) -> int:
        minute = _convert_quarterly_minutes(now_min, is_quarterly)
        ret = ThresholdBase.BASECURRENT
        if not charger_enabled or charger_done or moving_avg == 0:
            return ret
        currents = currents_dict
        for key, value in currents.items():
            if ((((moving_avg + key) / 60) * (60 - minute) + total_energy * 1000) / 1000) < peak:
                ret = value
                break
        return min(ret, power_canary_amp) if power_canary_amp > -1 else ret
